@@ -13,7 +13,7 @@ use strict;
 use warnings;
 use base qw(XML::API);
 
-our $VERSION = '0.13';
+our $VERSION = '0.14';
 
 use constant DOCTYPE => qq{<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">};
 
@@ -40,6 +40,10 @@ sub _root_attrs {
     return {xmlns => 'http://www.w3.org/1999/xhtml'};
 }
 
+sub _content_type {
+    return 'application/xhtml+xml';
+}
+
 
 1;
 
